/*
 *	Programme de Mandelbrot sequentiel  :
 *	cree un fichier pgm à visualiser avec eog
 *
 *	Compilation : gcc -o mandelbrot_seq mandelbrot_seq.c
 *
 *  Quelques jolies fractales
 *      ./mandel.ex 2000 2000 0.35 0.355 0.353 0.358 200
 *      ./mandel.ex 2000 2000 -0.736 -0.184 -0.735 -0.183 500
 *      ./mandel.ex 2000 2000 -0.736 -0.184 -0.735 -0.183 300
 *      ./mandel.ex 2000 2000 -1.48478 0.00006 -1.48440 0.00044 100
 */

#include <stdlib.h>
#include <stdio.h>
#include <time.h>	/* chronometrage */
#include "image.hpp"

#include <iostream>
#include <thread>
#include <vector>
#include "ProdCons.cpp"

using namespace std;

// *********************************************************************
//
//  Travail à faire, pour chaque ligne on calcule la couleur d'une ligne
//
// *********************************************************************
// description du travail
typedef struct {
  // les coordonnée de la zone à afficher
  double xmin;
  double xmax;
  double ymin;
  double ymax;
  // la dimension de l'image
  int dimX;
  int dimY;
  // la profondeur de calcul
  int prof;
} descr;

void usage(const char* nom) {
  printf("Usage : %s dim0 dim1 Xmin Ymin Xmax Ymax Prof \n", nom);
}

descr lecture_args(int argc, char *argv[]) {
  descr   d;

  /* Valeurs par defaut de la fractale */
  // 2000 2000 -0.736 -0.184 -0.735 -0.183 500
  d.xmin = -0.736;
  d.ymin = -0.184;
  d.xmax = -0.735;
  d.ymax = -0.183;
  d.dimX = 2000;
  d.dimY = 2000;
  d.prof = 500;
  
  /* Recuperation des parametres */
  if( argc > 1) {
    d.dimX = atoi(argv[1]);
    if (d.dimX == 0) {
      usage(argv[0]);
      exit(1);
    }
  }
  if ( argc > 2) {
    d.dimY = atoi(argv[2]);
    if (d.dimY == 0) {
      usage(argv[0]);
      exit(1);
    }
  }
  if (argc > 3) d.xmin = atof(argv[3]);
  if( argc > 4) d.ymin = atof(argv[4]);
  if( argc > 5) d.xmax = atof(argv[5]);
  if( argc > 6) d.ymax = atof(argv[6]);
  if( argc > 7)  {
    d.prof = atoi(argv[7]);
    if (d.prof == 0) {
      usage(argv[0]);
      exit(1);
    }
  }
  if (argc > 8) {
      usage(argv[0]);
      exit(1);
  }

  return d;
}

void calcul_ligne(image_t image, descr d, int numligne) {
  int i;
  int depart = numligne*d.dimY;
  double x = numligne*(d.xmax-d.xmin)/d.dimX+d.xmin;
  double y = d.ymin;
  double pas = (d.ymax-d.ymin)/d.dimY;

  if ((numligne<0)||(numligne>=d.dimX)) {
    fprintf(stderr, "Attention, vous tentez de calculer la ligne %d d'une image qui a %d lignes\n",
	    numligne, d.dimX);
    return;
  }
  
  pixel_t *p = image.data+depart;
  if (*p != 0) {
    fprintf(stderr, "Attention, la ligne %d de l'image image semble avoir déjà été calculée (je continue, mais il y a peut-être une erreur)\n",
	    numligne);    
  }
  for (i=0; i<d.dimY; i++) {
    p[i] = xy2color(x,y, d.prof);
    y += pas;
  }
}


// ********************************************************************************
// 
//   modifier à partir de là
//   Partie principale: pour chaque ligne de l'image appliquer la fonction précédante
// ********************************************************************************

void
f_thread(int numth, descr d, image_t &image, ProdCons<int> &pc, int &nblignes) {
  int numligne;

  fprintf(stdout, "lancement du thread %d\n", numth);
  nblignes = 0;

  while (1) {
    numligne = pc.get();
    if (numligne == -1) {
      break;
    }
    nblignes++;
    calcul_ligne(image, d, numligne);
  }
  fprintf(stdout, "fin du thread %d qui a calculé %d lignes\n", numth, nblignes);
}

int main(int argc, char *argv[]) {
  descr   d;
  int	  i;
  clock_t debut, fin;		    /* chronometrage */
  int nbth = 10;

  
  /* debut du chronometrage */
  debut = clock();

  /* lecture des argument */
  d = lecture_args(argc, argv);
  
  /* affichage parametres pour control */
  printf( "Domaine: %lf, %lf %lf %lf\n", 
	  d.xmin, d.ymin, d.xmax, d.ymax); 
  printf( "Prof: %d\n",  d.prof);
  printf( "Dim image: %d %d\n", d.dimX, d.dimY);
  
  // création de l'image
  image_t im = image_new(d.dimX, d.dimY);

  vector<thread> tabth;
  int tabnblignes[nbth];
  ProdCons<int> pc(20);
  for (i=0; i<nbth; i++) {
    tabth.push_back(thread(f_thread, i, d, std::ref(im), std::ref(pc), std::ref(tabnblignes[i])));
  }
    
  /* Traitement de l'image ligne par ligne*/
  for ( i = 0; i < d.dimX; i++) {
    pc.put(i);
  }

  for (i=0; i<nbth; i++) {
    pc.put(-1);
  }

  int nblignes = 0;
  for (i=0; i<nbth; i++) {
    tabth[i].join();
    nblignes += tabnblignes[i];
  }

  if (nblignes == d.dimX) {
    fprintf(stdout, "on a bien calculé %d lignes\n", nblignes);
  } else {
    fprintf(stdout, "Attention, %d lignes ont été calculées à la place de %d (ou  le calcul du nombre de ligne est faux)\n", nblignes, d.dimX);
  }

  
  /* Sauvegarde de la grille dans le fichier resultat "result.pgm" */
  image_to_file(im, "result.pgm");
  
  /* fin du chronometrage */
  fin = clock();
  printf("Temps total de calcul : %e seconde \n",
		 (double) (fin - debut)/ CLOCKS_PER_SEC);
  
  return 0;
}
